"""
Some usefull function.
"""

import numpy as np
import synumses.one_dimension.parameters as parameters


def calc_p_density():
    """
    Calculates and returns an array of the hole density.
    """
    p_density = parameters.Nv*np.exp(parameters.q*( parameters.u[1::3] - parameters.u[0::3] - parameters.Chi - parameters.Eg)/(parameters.kB*parameters.T))
          
    return p_density



def calc_n_density():
    """
    Calculates and returns an array of the electron density.
    """
    n_density = parameters.Nc*np.exp(parameters.q*( parameters.u[0::3] - parameters.u[2::3] + parameters.Chi)/(parameters.kB*parameters.T))
    
    return n_density


def ohm_potential(C, Chi, Eg, Nc, Nv):
   """
   Calculates and returns the potential from the doping level, the electron affinity, the band gap, and the density of states.
   """

   from synumses.one_dimension.parameters import q, kB, Ut, T
   
   ni = np.sqrt(Nc*Nv)*np.exp(-((Eg)*q)/(2.*kB*T))
   Phi = -Chi - Eg/2. - 0.5*Ut*np.log(Nc/Nv) + Ut*np.arcsinh(C/(2.*ni))
    
   return(Phi)


def calc_ni_density():
   """
    Calculates and returns an array of the intrinsic carrier density.
    """
   ni_density = (
      np.sqrt(parameters.Nc*parameters.Nv)
      *
      np.exp(-((parameters.Eg)*parameters.q)
             /
             (2.*parameters.kB*parameters.T))
   )

   return ni_density

def calc_recombination():
   """
   Calculates  and returns an array of the recombination rate.
   
   """
   recombination = (
      parameters.Cau
      *
      (calc_p_density() * calc_n_density() -
       calc_ni_density()**2
      )
   )
    
   return recombination

def calc_generation(I_0, alpha):
   """
   Does nothing so far.
   """
   
   return None
    



import setuptools

with open("DESCRIPTION.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(name='synumses-pkg-pabele',
                 version='0.8.4',
                 description='Numerical solver for semiconductor devices in Python',
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 url='',
                 author='Peter Abele',
                 author_email='ppabele@web.de',
                 license='GNU GPLv3',
                 packages=setuptools.find_packages(),
                 classifiers=[
                     "Development Status :: 4 - Beta",
                     "Programming Language :: Python :: 3.7",
                     "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
                     "Topic :: Scientific/Engineering",
                     "Topic :: Scientific/Engineering :: Mathematics",
                     "Topic :: Scientific/Engineering :: Physics",
                     "Topic :: Scientific/Engineering :: Visualization",
                     "Topic :: Software Development :: Libraries :: Python Modules",
                 ],
                 python_requires='>=3.7',
                 install_requires=[
                     'sympy',
                     'scipy',
                     'numpy',
                 ]

)



"""
Within this module all global variables and arrays are defined.
It also defines the default device, a silicon pn junction.

.. csv-table:: **global variables**
   :header: "global variable", "meaning", "default value"
   :widths: 10, 25, 10
   
   "T", "temperature", "300"
   "q", "elementary charge :math:`q_\mathrm{e}`", "1.6E-19"
   "kB", "Boltzman constant :math:`k_\mathrm{B}`",  "1.38E-23"
   "Ut", "temperature voltage :math:`U_\mathrm{T}`", "kB*T/q"
   "Epsilon_0", "vacuum permittivity :math:`\epsilon_0`", "8.85E-12"
   "Epsilon_r", "relative permittivity of silicon :math:`\epsilon_\mathrm{r}`",  "11.6"
   "n", "number of grid points", "400"
   "lx", "length of device", "400E-9"
   "bernoulli_limit", "for values less then **bernoulli_limit** the bernoulli function is approximated by a polynom", "1E1"

.. csv-table:: **global arrays**
   :header: "global array", "meaning", "default value"
   :widths: 10, 25, 10
   
   "pos_x", "position of grid point", "np.linspace(0,lx,n)"
  
   "Chi", "electron affinity", "np.full(n, 4.05)"
   "Eg", "band gap", "np.full(n, 1.12)"


   "Nc", "effective density of states in conduction band", "np.full(n, 2.81E25)"
   "Nv", "effective density of states in valence band", "np.full(n, 1.83E25)"


   "Epsilon", "permittivity of the material", "np.full(n, Epsilon_r * Epsilon_0)"
   "mu_p", "hole mobility :math:`\mu_\mathrm{p}`", "np.full(n, 0.045)"
   "mu_n", "electron mobility :math:`\mu_\mathrm{n}`", "np.full(n, 0.14)"
   "C", "doping", "np.zeros(n), -1E24/1E24"
   "Cau", "coefficient for recombination: q*(Cau*(n*p-ni2)-generation)*dx ", "np.full(n, 0)"
   "generation", "generation: q*(Cau*(n*p-ni2)-generation)*dx", "np.full(n, 0.0)"

   "b", ":math:`f(x)`", "np.zeros(3*n)"
   "A", "jacobian matrix",  "sparse.lil_matrix((3*n, 3*n))"
   "x", ":math:`\delta x`",  "np.zeros(3*n)"

"""


import numpy as np

import scipy.sparse as sparse 


# Global Variables

# #####################
#       Constants
# #####################
T = 300.
q = 1.6E-19

kB = 1.38E-23
Ut = kB*T/q

Epsilon_0 = 8.85E-12
Epsilon_r = 11.6

# ############
#    Mesh
# ############
n  = 400
lx = 400E-9


# *********************************************
#  Limit for Bernoulli function epproximation
# *********************************************

bernoulli_limit = 1E-4 # Default 1E-4


def init_geometry():
        """
        This functions sets grid size (**dx**) and the position of the gris points (**pos_x**).
        This function must be executed after changing the length (**lx**) or the number of grid points (**n**).    
        """
        

        global dx
        dx = lx/n

        global pos_x
        pos_x = np.linspace(0,lx,n)

def init_parameters():
        """
        This function defines the material parameters for silicon.
        After executing this function the parameters can be altered.
        This function must be executed after **init_geometry()**
        """

        global dt
        dt = 1E-12
        
        global Chi
        Chi = np.full(n, 4.05)

        global Eg
        Eg = np.full(n, 1.12)
        
        global Nc
        Nc = np.full(n, 2.81E25)

        global Nv
        Nv = np.full(n, 1.83E25)

        global Epsilon
        Epsilon = np.full(n, Epsilon_r * Epsilon_0)

        global mu_p
        mu_p = np.full(n, 0.045)

        global mu_n
        mu_n = np.full(n, 0.14)

        # Doping-Profile
        global C
        C = np.zeros(n)

        global Cau
        Cau = np.full(n, 0) # 1E-28
        
        global generation
        generation = np.full(n, 0.0)
 
        #
        global u
        u = np.zeros(3*n)

        #
        global u_old
        u_old = np.zeros(3*n)
        
        # 
        global b
        b = np.zeros(3*n)

        # 
        global A
        A = sparse.lil_matrix((3*n, 3*n))

        # Vector dx to be solved
        global x
        x = np.zeros(3*n)

        

def init_default_doping():
        """
        Definition of the default doping level.
        Left part doped with  :math:`N_\mathrm{a} = 1 \cdot 10^{24}\, \mathrm{m^{-3}}` and
        right part doped with :math:`N_\mathrm{d} = 1 \cdot 10^{24}\, \mathrm{m^{-3}}`
        
        """
        Na = 1E24
        Nd = 1E24

        for i in range(0,n):
                if i < n/2:
                    C[i] = -Na
                else:
                    C[i] = +Nd


        return None



init_geometry()
init_parameters()
init_default_doping()




"""
This module generates the python code for the
module **scharfetter_gummel_bernoulli.py**.
It solves the *Van Roosbroeck equations* using the 
*Scharfetter Gummel Scheme*.

It uses the SymPy module to generate the Jacobian matrix.
"""

from sympy import symbols, Symbol 
from sympy import Function
from sympy import solve, pretty
from sympy import diff
from sympy import sin, cos, exp, ln, sinh, cosh, sqrt, exp, Abs
from sympy import simplify, trigsimp
from sympy.parsing.sympy_parser import parse_expr

from sympy import srepr

import itertools

import time

#
# activate if simplify() is not to be used
# this results in much faster calculation
#
def simplify(x):
    return x

def bernoulli_poly(expr):
    """
    Returns a polynom to approximate the Bernoulli function. 
    """
    #return 1.0
    return 1.0 - expr/2.0
    #return 1.0 - expr/2.0 + expr**2/12.0 - expr**4/720.0 
    #return 1.0 - expr/2.0 + expr**2/12.0 - expr**4/720.0 + expr**6/30240.0 
    #return 1.0 - expr/2.0 + expr**2/12.0 - expr**4/720.0 + expr**6/30240.0 - expr**8/1209600.0
    #return expr/(exp(expr) - 1)


def bernoulli_exp(expr):
    """
    Returns the Bernoulli-function.
    """
    return expr/(exp(expr) - 1)
    
#
# Global variables
#
bernoulli = Function('bernoulli')
bernoulli_limit = symbols('parameters.bernoulli_limit')

x = Symbol('x')

Phi_n_k, Phi_n_l = symbols('Phi_n_k, Phi_n_l')
Phi_p_k, Phi_p_l = symbols('Phi_p_k, Phi_p_l')
Psi_k, Psi_l     = symbols('Psi_k, Psi_l')

Chi_k, Chi_l = symbols('Chi_k, Chi_l')
Eg_k, Eg_l   = symbols('Eg_k, Eg_l')
Nv_k, Nv_l   = symbols('Nv_k, Nv_l')
Nc_k, Nc_l   = symbols('Nc_k, Nc_l')

mu_n_k, mu_n_l   = symbols('mu_n_k, mu_n_l')
mu_p_k, mu_p_l   = symbols('mu_p_k, mu_p_l')

Psi_m1, Psi_00, Psi_p1       = symbols('parameters.u[3*(i-1)+0], parameters.u[(3*i+0)+0], parameters.u[3*(i+1)+0]')
Phi_p_m1, Phi_p_00, Phi_p_p1 = symbols('parameters.u[3*(i-1)+1], parameters.u[(3*i+0)+1], parameters.u[3*(i+1)+1]')
Phi_n_m1, Phi_n_00, Phi_n_p1 = symbols('parameters.u[3*(i-1)+2], parameters.u[(3*i+0)+2], parameters.u[3*(i+1)+2]')

Psi_old, Phi_p_old,  Phi_n_old = symbols('parameters.u_old[(3*i+0)+0], parameters.u_old[(3*i+0)+1], parameters.u_old[(3*i+0)+2]')

Chi_m1, Chi_00, Chi_p1    = symbols('parameters.Chi[i-1], parameters.Chi[i+0], parameters.Chi[i+1]')
Eg_m1, Eg_00, Eg_p1       = symbols('parameters.Eg[i-1], parameters.Eg[i+0], parameters.Eg[i+1]')

Nv_m1, Nv_00, Nv_p1       = symbols('parameters.Nv[i-1], parameters.Nv[i+0], parameters.Nv[i+1]')
Nc_m1, Nc_00, Nc_p1       = symbols('parameters.Nc[i-1], parameters.Nc[i+0], parameters.Nc[i+1]')

mu_n_m1, mu_n_00, mu_n_p1       = symbols('parameters.mu_n[i-1], parameters.mu_n[i+0], parameters.mu_n[i+1]')
mu_p_m1, mu_p_00, mu_p_p1       = symbols('parameters.mu_p[i-1], parameters.mu_p[i+0], parameters.mu_p[i+1]')

dx = symbols('parameters.dx')
dt = symbols('parameters.dt')

recombination = symbols('parameters.recombination[i]')
generation    = symbols('parameters.generation[i]')


C_00 = symbols('parameters.C[i]')
Epsilon_00 = symbols('parameters.Epsilon[i]')

np_exp  = Function('np.exp')
np_sqrt = Function('np.sqrt')
np_abs  = Function('np.abs')

kB = symbols('parameters.kB')
T  = symbols('parameters.T')
q  = symbols('parameters.q')


Ut = (kB * T) / q


Cau = symbols('parameters.Cau[i]')

#ni = Function('ni')
#p = Function('p')
#n = Function('n')

ni2 =(Nv_00*Nc_00)*exp(-Eg_00/Ut)
p  = Nv_00*exp(( (-Chi_00 - Eg_00) + Phi_p_00 - Psi_00)/Ut)
n  = Nc_00*exp((        +Chi_00    - Phi_n_00 + Psi_00)/Ut)

p_old  = Nv_00*exp(( (-Chi_00 - Eg_00) + Phi_p_old - Psi_old)/Ut)
n_old  = Nc_00*exp((        +Chi_00    - Phi_n_old + Psi_old)/Ut)


# **********************************
# **********************************
#     Van Roosbroeck equations
# **********************************
# **********************************

#
# Poisson equaton
#
poisson = ((Psi_p1 - 2.*Psi_00 + Psi_m1) + q / Epsilon_00 * (C_00 + p - n)*dx**2)


# 
# Electrical hole current density
#
j_p =+kB*T/dx*(mu_p_k + mu_p_l)/2.*(Nv_k + Nv_l)/2.*exp(-(Chi_k + Eg_k + Chi_l + Eg_l)/(2.*Ut))*(
    +bernoulli(+(Psi_l-Psi_k)/(Ut)) * exp((Phi_p_k - Psi_k)/(Ut))
    -bernoulli(-(Psi_l-Psi_k)/(Ut)) * exp((Phi_p_l - Psi_l)/(Ut))
)


j_n = -kB*T/dx*(mu_n_k + mu_n_l)/2.*(Nc_k + Nc_l)/2.*exp((Chi_k + Chi_l)/(2.*Ut))*(
    +bernoulli(-(Psi_l-Psi_k)/(Ut)) * exp((Psi_k - Phi_n_k)/Ut)
    -bernoulli(+(Psi_l-Psi_k)/(Ut)) * exp((Psi_l - Phi_n_l)/Ut)
)

#
# Divergence of current density
# 
div_j_p = (
    - j_p.subs([(Psi_k, Psi_00),(Psi_l, Psi_p1),(Phi_p_k, Phi_p_00),(Phi_p_l,Phi_p_p1),(Eg_k, Eg_00),(Eg_l, Eg_p1),
                (Chi_k, Chi_00),(Chi_l, Chi_p1),(Nv_k, Nv_00),(Nv_l, Nv_p1),(mu_p_k, mu_p_00),(mu_p_l, mu_p_p1)])
    + j_p.subs([(Psi_k, Psi_m1),(Psi_l, Psi_00),(Phi_p_k, Phi_p_m1),(Phi_p_l,Phi_p_00),(Eg_k, Eg_m1),(Eg_l, Eg_00),
                (Chi_k, Chi_m1),(Chi_l, Chi_00),(Nv_k, Nv_m1),(Nv_l, Nv_00),(mu_p_k, mu_p_m1),(mu_p_l, mu_p_00)])
    - q*(Cau*(n*p-ni2)-generation)*dx
)

div_j_n = (
    + j_n.subs([(Psi_k, Psi_00),(Psi_l, Psi_p1),(Phi_n_k, Phi_n_00),(Phi_n_l,Phi_n_p1),
                (Chi_k, Chi_00),(Chi_l, Chi_p1),(Nc_k, Nc_00),(Nc_l, Nc_p1),(mu_n_k, mu_n_00),(mu_n_l, mu_n_p1)])
    - j_n.subs([(Psi_k, Psi_m1),(Psi_l, Psi_00),(Phi_n_k, Phi_n_m1),(Phi_n_l,Phi_n_00),
                (Chi_k, Chi_m1),(Chi_l, Chi_00),(Nc_k, Nc_m1),(Nc_l, Nc_00),(mu_n_k, mu_n_m1),(mu_n_l, mu_n_00)])
    - q*(Cau*(n*p-ni2)-generation)*dx
)

#
# Time dependent divergence of current density
#
div_j_p_t = div_j_p + q*(p - p_old)*(dx/dt)
div_j_n_t = div_j_n + q*(n - n_old)*(dx/dt)

#
# The variable functions is defined before calling the functions
#                   makeUpdate_b and makeJacobi
# 
functions = []

partial_derivatives = [
    ["3*(i-1)+0",Psi_m1  ],
    ["3*(i+0)+0",Psi_00  ],
    ["3*(i+1)+0",Psi_p1  ],
    ["3*(i-1)+1",Phi_p_m1],
    ["3*(i+0)+1",Phi_p_00],
    ["3*(i+1)+1",Phi_p_p1],
    ["3*(i-1)+2",Phi_n_m1],
    ["3*(i+0)+2",Phi_n_00],
    ["3*(i+1)+2",Phi_n_p1]
]


substitutes = {
    "left"  : {
        Psi_m1: Symbol('(ohm_potential(parameters.C[0], parameters.Chi[0], parameters.Eg[0], parameters.Nc[0], parameters.Nv[0]) + Ua)'),
        Phi_p_m1: Symbol('Ua'),
        Phi_n_m1: Symbol('Ua'),

        Chi_m1: Chi_00,
        Eg_m1: Eg_00,
        Nv_m1: Nv_00,
        Nc_m1: Nc_00,
        mu_n_m1: mu_n_00,
        mu_p_m1: mu_p_00,
        
        exp:  np_exp,
        sqrt: np_sqrt,
        Abs:  np_abs
    },
    "right" : {
        Psi_p1: Symbol('(ohm_potential(parameters.C[parameters.n-1], parameters.Chi[parameters.n-1], parameters.Eg[parameters.n-1], parameters.Nc[parameters.n-1], parameters.Nv[parameters.n-1]) + Ub)'),
        Phi_p_p1: Symbol('Ub'),
        Phi_n_p1: Symbol('Ub'),
        
        Chi_p1: Chi_00,
        Eg_p1: Eg_00,
        Nv_p1: Nv_00,
        Nc_p1: Nc_00,
        mu_n_p1: mu_n_00,
        mu_p_p1: mu_p_00,
        
        exp:  np_exp,
        sqrt: np_sqrt,
        Abs:  np_abs
    },
    "center": {
        exp:  np_exp,
        sqrt: np_sqrt,
        Abs:  np_abs
    }
}

#
# Depending on the value of the argument the Bernoulli function is substituded
#
search_sub_function = ({"function"  : bernoulli,
                        "if_states" : ['<','<'],
                        "if_values" : [-bernoulli_limit, bernoulli_limit],
                        "if_subs"   : [bernoulli_exp , bernoulli_poly],
                        "else_sub"  : bernoulli_exp
})


def substituteFunctions(function, sub_function, left_side, sub_args = [], partial_derivative = None, tabs=0, do_simplify = 0):
    """
    Depending on the value of the argument of a function it is substituted.
    """

#    print("### function:", function)
#    print("### sub args:", sub_args)
#    print("### left_side:", left_side)
#    print("### sub_args", sub_args)
#    print()
#    print("### function.subs(sub_args):", function.subs(sub_args))
#    if partial_derivative  is not None:
#        print("### diff(function, partial_derivative).subs(sub_args):", diff(function, partial_derivative).subs(sub_args))
#    print("### srepr(function):", srepr(function))
#    print("### sub_function:", sub_function)


# ****************************************************
# Search for all arguments of sub_function["function"]
# ****************************************************
    
    comb = []
    found_arguments = []
    
    for funcs in function.find(Function): #loop all functions
        if(str(type(funcs)) == str(sub_function["function"])):
                found_arguments.append(funcs.args[0])

    comb = (sub_function, found_arguments)

    perm = list(itertools.product(range(len(sub_function["if_states"])+1), repeat = len(found_arguments)))


    
    if not found_arguments:

        for ii in range(tabs):
            print("\t", end = '')

        print(left_side, "= ", end = '')

        
        if  partial_derivative is not None:
                
            part_der = diff(simplify(function), partial_derivative).subs(sub_args)
            
            if (do_simplify):
                print(simplify(part_der))
            else:
                print(part_der)
        else:

            if (do_simplify):
                print(simplify(function).subs(sub_args))
            else:
                print(function.subs(sub_args))


    else:
        
        for i in range(len(perm)):

            for ii in range(tabs):
                print("\t", end = '')

            if (i==0):
                print("if (", end = '')
            elif(i<(len(perm)-1)):
                print("elif (", end = '')
            else:
                print("else", end = '')

            stateSet = 0;
            substitutes = []
            for j in range(len(found_arguments)): #number of arguments

                #print("comb[0][]:", comb[0]["function"](comb[1][j]))
                #print("comb[0][]:", comb[0]["if_subs"][perm[i][j]](comb[1][j]))

                #substitutes.append((comb[0]["function"](comb[1][j]), comb[0]["if_subs"][perm[i][j]](comb[1][j])))

                if (j==0):
                    if (perm[i][j] != len(sub_function["if_states"])):
                        #print("(perm:",perm[i][j], "arg:", j, ")",end = '')
                        print("(", comb[1][j].subs(sub_args), comb[0]["if_states"][perm[i][j]], comb[0]["if_values"][perm[i][j]], ")", end = '')
                        substitutes.append((comb[0]["function"](comb[1][j]), comb[0]["if_subs"][perm[i][j]](comb[1][j])))
                        stateSet=1
                    else:
                        substitutes.append((comb[0]["function"](comb[1][j]), comb[0]["else_sub"](comb[1][j])))
                        
                else:
                    if (perm[i][j] != len(sub_function["if_states"])):
                        substitutes.append((comb[0]["function"](comb[1][j]), comb[0]["if_subs"][perm[i][j]](comb[1][j])))
                        if (stateSet==1):
                            #print(" and ", "(perm:", perm[i][j], "arg:", j, ")",end = '')
                            print(" and (", comb[1][j].subs(sub_args), comb[0]["if_states"][perm[i][j]], comb[0]["if_values"][perm[i][j]], ")", end = '')
                            
                        else:
                           #print("(perm:", perm[i][j], "arg:", j, ")",end = '') 
                           print("(", comb[1][j].subs(sub_args), comb[0]["if_states"][perm[i][j]], comb[0]["if_values"][perm[i][j]], ")", end = '')
                           stateSet = 1;
                    else:
                       substitutes.append((comb[0]["function"](comb[1][j]), comb[0]["else_sub"](comb[1][j]))) 
                


            if(i<(len(perm)-1)):
                print("):\n", end = '')
            else:
                print(":\n", end = '')

            for ii in range(tabs):
                print("\t", end = '')

            print("\t" , end = '')
            print(left_side, "= ", end = '')

            if  partial_derivative is not None:
                
                part_der = diff(function.subs(substitutes), partial_derivative).subs(sub_args)

                if (do_simplify):
                    print(simplify(part_der))
                else:
                    print(part_der)
            else:

                if (do_simplify):
                    print(simplify(function).subs(substitutes).subs(sub_args))
                else:
                    print(function.subs(substitutes).subs(sub_args))


            print()


def makeUpdate_b(name):

    """
    Generates the function vector.
    """

    print("################################")
    print("########### {name} ###########".format(name=name))
    print("################################")

    print("def {name}(Ua, Ub):".format(name=name)) 
    print()
    print("\tfor i in range(0, parameters.n):")


    for s in ["left", "right", "center"]:

        if (s =="left"):
            print("\t\tif (i==0) :")
        elif (s=="right"):
            print("\t\telif(i==parameters.n-1):")
        else:
            print("\t\telse:")

        print("\t\t\t#################")
        print("\t\t\t### ",s, "###")
        print("\t\t\t#################")
        for function in functions:

             substituteFunctions(function[1],
                                 search_sub_function,
                                 left_side = "parameters.b["+function[0]+"]",
                                 sub_args = substitutes[s],
                                 tabs = 3,
                                 do_simplify = 1
                )

    print() 
    print("\treturn None")
    print()
    print()
    print()

    
def makeJacobi(name):
    
    print("######################################")
    print("###########  {name} ##################".format(name=name))
    print("######################################")

    print("def {name}(Ua, Ub):".format(name=name))
    print()
    print("\tfor i in range(0, parameters.n):")

    for s in ["left", "right", "center"]:

        if (s =="left"):
            print("\t\tif (i==0) :")
        elif (s=="right"):
            print("\t\telif(i==parameters.n-1):")
        else:
            print("\t\telse:")

        print()
        print("\t\t\t#################")
        print("\t\t\t### ",s, "###")
        print("\t\t\t#################")

        #print("# ",substitutes[s])

        for function in functions:
            print
            print("\t\t\t#######################")
            print("\t\t\t### ",function[2]," ###")
            print("\t\t\t#######################")
            for partial_derivative in partial_derivatives:
                if ((s == "left"  and (partial_derivative[1] in [Psi_m1, Phi_p_m1, Phi_n_m1])) or
                    (s == "right" and (partial_derivative[1] in [Psi_p1, Phi_p_p1, Phi_n_p1]))):
                    print("\t\t\t###"+"parameters.A["
                          +function[0]
                          +","
                          +partial_derivative[0]
                          +"]",
                          " = ..."
                    )
                else:
                    substituteFunctions(-function[1],
                                        search_sub_function, 
                                        "parameters.A["+function[0]+","+partial_derivative[0]+"]",
                                        sub_args = substitutes[s],
                                        partial_derivative = partial_derivative[1],
                                        tabs=4,
                                        do_simplify = 1)

    print()
    print("\treturn ")
    print()
    print()

    
    
def codeGenerator():
    """
    This function generates the python code. 
    """
    # ******************************************************************
    #  functions is a global variable, being modified in this function
    # ******************************************************************
    global functions
    
   
    print("####################################")
    print("### This code was automatically  ###")
    print("###        generated             ###")
    print("###    on",time.strftime("%d.%m.%Y at %H:%M"), "   ###")
    print("####################################")
   
    print()
    
    print("#################################")
    print("###########  Import   ###########")
    print("#################################")
    print("import numpy as np")
    print()
    print("import synumses.one_dimension.parameters as parameters")
    print("from   synumses.one_dimension.functions import ohm_potential")

    print()
    print()

    print("#################################")
    print("########### Bernoulli ###########")
    print("#################################")
    print("def bernoulli(x):")

    substituteFunctions(bernoulli(x),
                        search_sub_function,
                        "x",
                        sub_args = {
                            exp: np_exp},
                        tabs = 1,
                        do_simplify = 1
    )
    print("\treturn(x)")
    
    print("############################################")
    print("########### hole_current_density ###########")
    print("############################################")
    print("def hole_current_density():")
    print()
    print("\tj_p = np.zeros(parameters.n)")
    print()
    print("\tfor i in range(0,parameters.n-1):")

    substituteFunctions(j_p,
                        search_sub_function,
                        "j_p[i]",
                        sub_args = {
                            Psi_k:   Psi_00, 
                            Psi_l:   Psi_p1,
                            Phi_p_k: Phi_p_00,
                            Phi_p_l: Phi_p_p1,
                            Chi_k:   Chi_00,
                            Chi_l:   Chi_p1,
                            Eg_k:    Eg_00,
                            Eg_l:    Eg_p1,
                            Nv_k:    Nv_00,
                            Nv_l:    Nv_p1,
                            mu_p_k:  mu_p_00,
                            mu_p_l:  mu_p_p1,
                            mu_n_k:  mu_n_00,
                            mu_n_l:  mu_n_p1,
                            exp:     np_exp,
                            Abs:     np_abs
                        },
                        tabs = 2,
                        do_simplify = 1
    )

    print()
    print("\ti = parameters.n-1")
    print("\tj_p[i] =  j_p[i-1]")
    print()
    print("\treturn j_p")
    print()
    print()

    print("################################################")
    print("########### electron_current_density ###########")
    print("################################################")
    print("def electron_current_density():")
    print()
    print("\tj_n = np.zeros(parameters.n)")
    print()
    print("\tfor i in range(0,parameters.n-1):")

    substituteFunctions(j_n,
                        search_sub_function,
                        "j_n[i]",
                        sub_args = {
                            Psi_k:   Psi_00, 
                            Psi_l:   Psi_p1,
                            Phi_n_k: Phi_n_00,
                            Phi_n_l: Phi_n_p1,
                            Chi_k:   Chi_00,
                            Chi_l:   Chi_p1,
                            Nc_k:    Nc_00,
                            Nc_l:    Nc_p1,
                            mu_p_k:  mu_p_00,
                            mu_p_l:  mu_p_p1,
                            mu_n_k:  mu_n_00,
                            mu_n_l:  mu_n_p1,
                            exp:     np_exp,
                            Abs:     np_abs
                        },
                        tabs = 2,
                        do_simplify = 1
    )



    print()
    print("\ti = parameters.n-1")
    print("\tj_n[i] =  j_n[i-1]")
    print()
    print("\treturn j_n")
    print()
    print()

    # *****************************************************
    #
    # Generating the Jacobian matrix and b with biasing
    #
    # *****************************************************

    functions = [
        ["3*i+0",poisson, "poisson"],
        ["3*i+1",div_j_p, "div_j_p"],
        ["3*i+2",div_j_n, "div_j_n"]
    ]
    makeUpdate_b("update_b")
    makeJacobi("jacobian")

    # *****************************************************
    #
    # Generating the Jacobian matrix and b without biasing
    # for constant Fermi level 
    #
    # *****************************************************

    functions = [
        ["3*i+0",poisson, "poisson"],
        ["3*i+1",Phi_p_00 - Phi_p_m1, "div_j_p"], #Difference to the not first equations, results in a constant quasi Fermi level
        ["3*i+2",Phi_n_00 - Phi_n_m1, "div_j_n"]  #Difference to the not first equations, results in a constant quasi Fermi level
    ]
    makeUpdate_b("first_update_b")
    makeJacobi("first_jacobian")

    
    # *****************************************************
    #
    # Generating the Jacobian matrix and b for transient
    # for constant Fermi level 
    #
    # *****************************************************


    #
    # As div_j_p and div_j_n have changed functions must be updated
    #
    functions = [
        ["3*i+0",poisson, "poisson"],
        ["3*i+1",div_j_p_t, "div_j_p"],
        ["3*i+2",div_j_n_t, "div_j_n"]
    ]

    makeUpdate_b("transient_update_b")
    makeJacobi("transient_jacobian")
    

if __name__ == '__main__':

   codeGenerator()



"""
This module contains the submodules solve one dimensional electronic devices.
"""

from .parameters import *
from .functions import *
from .solver import *
from .scharfetter_gummel_bernoulli import *
from .python_code_generator import *



"""
SyNumSeS
========

SyNumSeS is a Python package for simulating various semiconductor devices
in one dimension.
It solves the the *Van Roosbroeck equations* using the *Scharfetter Gummel Discretization*.

"""


#import synumses.one_dimension
from .one_dimension import *



import scipy.sparse as sparse 
from scipy.sparse.linalg import spsolve, qmr

import numpy as np

from synumses.one_dimension.scharfetter_gummel_bernoulli import bernoulli, jacobian, update_b, first_jacobian, first_update_b, transient_jacobian, transient_update_b

import synumses.one_dimension.parameters as parameters

from synumses.one_dimension.functions import calc_recombination, calc_p_density, calc_n_density, ohm_potential


def find_nearest(a, a0):
    "Index in nd array `a` closest to the scalar value `a0`"
    idx = np.abs(a - a0).argmin()
    return idx

def time_step():
    
    parameters.u_old =  parameters.u


def solve_from_doping():
    
    for i in range(0,parameters.n):
            parameters.u[3*i + 0] = ohm_potential(parameters.C[i], parameters.Chi[i], parameters.Eg[i], parameters.Nc[i], parameters.Nv[i])
            parameters.u[3*i + 1] = 0.
            parameters.u[3*i + 2] = 0.
    return None



def solve_no_bias(norm_max = 1E-14, info = 'no'):
    
    solve_from_doping()
    
    while True:

        
        first_jacobian(0.0, 0.0)
        first_update_b(0.0, 0.0)

        calc_recombination()
        

        try:
            parameters.x = spsolve(parameters.A, parameters.b, use_umfpack=True)
        except RuntimeError:
            return False
        
        parameters.u = parameters.u + parameters.x
    
        norm = np.linalg.norm(parameters.x)

        if (info != 'no'):
            print("Norm of b: ", np.linalg.norm(parameters.b))
            print("Norm of x:" , np.linalg.norm(parameters.x))
        

        if (norm < norm_max):
            break

        
        #calc_p_density()
        #calc_n_density()
       
        
    return True


def solve_bias(Ua, Ub, norm_max  = 1E-6, info = 'no'):

    counter = 0

    while True:
        
        jacobian(Ua, Ub) 
        update_b(Ua, Ub)

        try:
            parameters.x = spsolve(parameters.A, parameters.b, use_umfpack=True)
        except RuntimeError:
            return False

        #parameters.x = qmr(parameters.A, parameters.b)
        parameters.u = parameters.u + parameters.x
    
        norm = np.linalg.norm(parameters.x)

        if (info != 'no'):
            print("Norm of b: ", np.linalg.norm(parameters.b))
            print("Norm of x:" , np.linalg.norm(parameters.x))

        #break
        

        if (np.isnan(norm) or (counter==20)):
            print("norm is nan")
            return False

        if (norm < norm_max):
            break

        counter = counter +1
  
    return True


def solve_bias_transient(Ua, Ub, dt, norm_max  = 1E-6, info = 'no'):
    
    parameters.dt = dt
    
    counter = 0

    while True: 
        
        transient_jacobian(Ua, Ub) 
        transient_update_b(Ua, Ub)

        try:
            parameters.x = spsolve(parameters.A, parameters.b, use_umfpack=True)
        except RuntimeError:
            return False

        #parameters.x = qmr(parameters.A, parameters.b)
        parameters.u = parameters.u + parameters.x
    
        norm = np.linalg.norm(parameters.x)

        if (info != 'no'):
            print("Norm of b: ", np.linalg.norm(parameters.b))
            print("Norm of x:" , np.linalg.norm(parameters.x))

        #break
        

        if (np.isnan(norm) or (counter==20)):
            print("norm is nan")
            return False

        if (norm < norm_max):
            break

        counter = counter +1


  
    return True

    
def solve_bias_center_boundary(Ua, Ub, Uc, Pos, doping, norm_max  = 1E-10, info = 'no'):
    
    prev_norm = 1E30

    while True:
    
        #calc_recombination()     
        
        jacobian(Ua, Ub) 
        update_b(Ua, Ub)
    
        base_pos =  find_nearest(parameters.pos_x, Pos)

        #
        # Boundary condition for potential
        #
        
        parameters.b[3*(base_pos+0)+0] =  ohm_potential(parameters.C[base_pos],
                                                        parameters.Chi[base_pos],
                                                        parameters.Eg[base_pos],
                                                        parameters.Nc[base_pos],
                                                        parameters.Nv[base_pos]) + Uc - parameters.u[3*(base_pos+0)+0] 

        parameters.A[3*(base_pos+0)+0, 3*(base_pos-1)+0] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos-1)+1] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos-1)+2] = 0

        parameters.A[3*(base_pos+0)+0, 3*(base_pos+0)+0] = 1
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+0)+1] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+0)+2] = 0

        parameters.A[3*(base_pos+0)+0, 3*(base_pos+1)+0] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+1)+1] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+1)+2] = 0

        if (doping =="p"):
            #
            # Boundary condition for hole fermi level
            #
            parameters.b[3*(base_pos+0)+1] =  Uc - parameters.u[3*(base_pos+0)+1] 

            parameters.A[3*(base_pos+0)+1, 3*(base_pos-1)+0] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos-1)+1] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos-1)+2] = 0

            parameters.A[3*(base_pos+0)+1, 3*(base_pos+0)+0] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+0)+1] = 1
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+0)+2] = 0

            parameters.A[3*(base_pos+0)+1, 3*(base_pos+1)+0] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+1)+1] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+1)+2] = 0

        elif (doping =="n"):
            #
            # Boundary condition for electron fermi level
            #
            parameters.b[3*(base_pos+0)+2] =  Uc - parameters.u[3*(base_pos+0)+2] 

            parameters.A[3*(base_pos+0)+2, 3*(base_pos-1)+0] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos-1)+1] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos-1)+2] = 0

            parameters.A[3*(base_pos+0)+2, 3*(base_pos+0)+0] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+0)+1] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+0)+2] = 1

            parameters.A[3*(base_pos+0)+2, 3*(base_pos+1)+0] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+1)+1] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+1)+2] = 0

        try:
            parameters.x = spsolve(parameters.A, parameters.b, use_umfpack=True)
        except RuntimeError:
            return False

        #parameters.x = qmr(parameters.A, parameters.b)
        parameters.u = parameters.u + parameters.x
    
        norm = np.linalg.norm(parameters.x)

        if (info != 'no'):
            print("Norm of b: ", np.linalg.norm(parameters.b))
            print("Norm of x:" , np.linalg.norm(parameters.x))
        
        if (np.isnan(norm)):
            print("norm is nan")
            return False

        if ((prev_norm <= norm) and (norm < norm_max)):
            break
             
        prev_norm = norm
  
    return True


def solve_bias_center_boundary_transient(Ua, Ub, Uc, Pos, doping, dt, norm_max  = 1E-10, info = 'no'):

    prev_norm = 1E30

    parameters.dt = dt
    
    while True:
    
        #calc_recombination()     
        
        transient_jacobian(Ua, Ub) 
        transient_update_b(Ua, Ub)
    
        base_pos =  find_nearest(parameters.pos_x, Pos)

        #
        # Boundary condition for potential
        #
        
        parameters.b[3*(base_pos+0)+0] =  ohm_potential(parameters.C[base_pos],
                                                        parameters.Chi[base_pos],
                                                        parameters.Eg[base_pos],
                                                        parameters.Nc[base_pos],
                                                        parameters.Nv[base_pos]) + Uc - parameters.u[3*(base_pos+0)+0] 

        parameters.A[3*(base_pos+0)+0, 3*(base_pos-1)+0] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos-1)+1] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos-1)+2] = 0

        parameters.A[3*(base_pos+0)+0, 3*(base_pos+0)+0] = 1
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+0)+1] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+0)+2] = 0

        parameters.A[3*(base_pos+0)+0, 3*(base_pos+1)+0] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+1)+1] = 0
        parameters.A[3*(base_pos+0)+0, 3*(base_pos+1)+2] = 0

        if (doping =="p"):
            #
            # Boundary condition for hole fermi level
            #
            parameters.b[3*(base_pos+0)+1] =  Uc - parameters.u[3*(base_pos+0)+1] 

            parameters.A[3*(base_pos+0)+1, 3*(base_pos-1)+0] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos-1)+1] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos-1)+2] = 0

            parameters.A[3*(base_pos+0)+1, 3*(base_pos+0)+0] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+0)+1] = 1
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+0)+2] = 0

            parameters.A[3*(base_pos+0)+1, 3*(base_pos+1)+0] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+1)+1] = 0
            parameters.A[3*(base_pos+0)+1, 3*(base_pos+1)+2] = 0

        elif (doping =="n"):
            #
            # Boundary condition for electron fermi level
            #
            parameters.b[3*(base_pos+0)+2] =  Uc - parameters.u[3*(base_pos+0)+2] 

            parameters.A[3*(base_pos+0)+2, 3*(base_pos-1)+0] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos-1)+1] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos-1)+2] = 0

            parameters.A[3*(base_pos+0)+2, 3*(base_pos+0)+0] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+0)+1] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+0)+2] = 1

            parameters.A[3*(base_pos+0)+2, 3*(base_pos+1)+0] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+1)+1] = 0
            parameters.A[3*(base_pos+0)+2, 3*(base_pos+1)+2] = 0

        try:
            parameters.x = spsolve(parameters.A, parameters.b, use_umfpack=True)
        except RuntimeError:
            return False

        #parameters.x = qmr(parameters.A, parameters.b)
        parameters.u = parameters.u + parameters.x
    
        norm = np.linalg.norm(parameters.x)

        if (info != 'no'):
            print("Norm of b: ", np.linalg.norm(parameters.b))
            print("Norm of x:" , np.linalg.norm(parameters.x))
        
        if (np.isnan(norm)):
            print("norm is nan")
            return False

        if ((prev_norm <= norm) and (norm < norm_max)):
            break
             
        prev_norm = norm
  
    return True


